          

(defparameter *start-symbol* 's)

(defparameter *grammar*
  (list   
;; Grammar rules
     (make-rule :lhs 's :rhs '(np vp))    
     (make-rule :lhs 'vp :rhs '(iv)) 
     (make-rule :lhs 'vp :rhs '(iv pp)) 
     (make-rule :lhs 'vp :rhs '(tv np)) 
     (make-rule :lhs 'vp :rhs '(tv np vp))
     (make-rule :lhs 'np :rhs '(det n)) 
     (make-rule :lhs 'np :rhs '(det n pp)) 
     (make-rule :lhs 'pp :rhs '(p np)) 
;; Lexicon 
     (make-rule :lhs 'det :rhs '(the)) 
     (make-rule :lhs 'det :rhs '(her)) 
     (make-rule :lhs 'np :rhs '(her)) 
     (make-rule :lhs 'np :rhs '(they)) 
     (make-rule :lhs 'n :rhs '(nurses))
     (make-rule :lhs 'tv :rhs '(nurses)) 
     (make-rule :lhs 'n :rhs '(book))
     (make-rule :lhs 'tv :rhs '(book))
     (make-rule :lhs 'n :rhs '(travel))
     (make-rule :lhs 'tv :rhs '(travel))  
     (make-rule :lhs 'n :rhs '(report))   
     (make-rule :lhs 'tv :rhs '(report)) 
     (make-rule :lhs 'tv :rhs '(hear)) 
     (make-rule :lhs 'tv :rhs '(see)) 
     (make-rule :lhs 'p  :rhs '(on))))













;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Edge data type 
;; (Version without extra component used for "packing")
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct edge from to category daughters unanalyzed)

(defun passive-edge-p (edge)
  (null (edge-unanalyzed edge)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rule data type
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct rule lhs rhs)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;;    Implementasjon av fundamentalregelen egnet til bruk i oblig2.
;;    Returnerer listen av kanter som kan avledes fra to vilkrlige
;;    kanter edge1 og edge2 i ett trinn ved hjelp av fundamentalregelen.  
;;    Denne listen har aldri mer enn ett element. Null elementer svarer til
;;    tilfeller der fundamentalregelen  IKKE kan anvendes p de to kantene.
;;    Denne prosedyren setter ingen krav til rekkeflgen p argumentene.
;;    Inneholder to kall p tilsvarende funksjon som forventer den aktive
;;    kanten frst og den passive etterp.

(defun fundamental-rule-unordered (edge1 edge2)
  (append (fundamental-rule edge1 edge2)
          (fundamental-rule edge2 edge1)))

(defun fundamental-rule (active passive)
  (if (and (not (passive-edge-p active))
           (passive-edge-p passive)
           (eq (edge-to active) (edge-from passive))
           (eq (first (edge-unanalyzed active)) (edge-category passive)))
      (list  (make-edge :from (edge-from active) 
                        :to (edge-to passive)
                        :category (edge-category active)
                        :daughters (append (edge-daughters active) (list (first (edge-unanalyzed active))))  
                        :unanalyzed (rest (edge-unanalyzed active))))
      (list)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Implementasjon av bu-prediksjonsregelen.  
;;  For hver passiv kant med kategori cat setter denne regelen igang en ny aktiv kant 
;;  for hver produksjon i grammatikken med cat frst p hyresiden.
;;  Frst en hjelperutine som plukker ut de relevante reglene:

(defun expect-first (cat rules)
    (if (null rules)
        (list)
      (if (eq cat (first (rule-rhs (first rules))))
          (cons (first rules) (expect-first cat (rest rules)))
        (expect-first cat (rest rules)))))

;;  S en som lager ny aktiv kant svarende til en regel:

(defun startup (rule from to)
    (make-edge :from from
               :to to
               :category (rule-lhs rule)
               :daughters (list (first (rule-rhs rule)))
               :unanalyzed (rest (rule-rhs rule))))

;;  Og til slutt funksjonen selv:

(defun bu-prediction (edge)
    (if (passive-edge-p edge)
        (loop for rule in (expect-first (edge-category edge) *grammar*)
              collect (startup rule (edge-from edge) (edge-to edge)))
      (list)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Implementasjon av nullre regler, alts regler som gir oss kanter
;; vi kan starte opp, det vil si kanter som ikke er avledet fra andre kanter.
;; Disse tar alts ingen kantargumenter, men det kan vre naturlig  gi dem
;; inputstrengen som argument.
;;     
;;   1) Implementasjon av top-down-oppstartsregelen, som setter igang
;;      aktive kanter som leter etter en S som begynner frst i strengen.
;;
;;   Hjelpefunksjon som starter ny, aktiv kant i angitt posisjon:
  
(defun start-empty-edge (rule position)
  (make-edge :from position
             :to position
             :category (rule-lhs rule)
             :daughters (list)
             :unanalyzed (rule-rhs rule)))

;;   Hjelpefunksjon som finner regler med startsymbolet til venstre

(defun expand-cat (cat rules)
    (if (null rules)
        (list)
      (if (eq cat (rule-lhs (first rules)))
          (cons (first rules) (expand-cat cat (rest rules)))
        (expand-cat cat (rest rules)))))

(defparameter *expand-startsymbol* (expand-cat *start-symbol* *grammar*))

;;  Og her kommer den.  Legg merke til at den ikke gjr bruk av  argumentet sitt.

(defun td-start (inputsentence)
  (loop for rule in *expand-startsymbol*
      collect (start-empty-edge rule 0)))

;;   2)  Implementasjon av initieringsregel som lager en passiv kant for hvert
;;       ord i input.

(defun readwords (input)
  (loop  for i from 0
         for word in input
         collect (make-edge :from i :to (+ i 1) :category word)))






